/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * VariantsHunter.java
 *
 * Created on 12 gennaio 2003, 11.56
 * @author  Valter Crescenzi
 */


package roadrunner.engine;

import java.util.*;
import java.util.logging.*;

import roadrunner.util.Util;
import roadrunner.ast.Expression;
import roadrunner.ast.Walker;
import roadrunner.ast.Node;
import roadrunner.ast.ASTBuilder;
import roadrunner.ast.ASTVariant;
import roadrunner.ast.ASTToken;
import roadrunner.parser.Token;
import roadrunner.parser.Binder;
import roadrunner.parser.TokenList;
import roadrunner.parser.BindingException;

/**
 * Look for variants, that is, pcdata, link or images
 * which match with at least two different values.
 * Collect samples of each variant.
 */
public class VariantsHunter {
    
    static private Logger log = Logger.getLogger(VariantsHunter.class.getName());
    
    private Set expressions;
    private Set ambiguous;
    private TokenList[] tokenlists;
    
    public VariantsHunter(Set expressions, TokenList[] tokenlists) {
        this.expressions = expressions;
        this.tokenlists = tokenlists;
        this.ambiguous = null;
    }
    /**
     * @return the list of ambiguous expressions
     **/
    public Set insertVariants() {
        Set result = new LinkedHashSet();
        this.ambiguous = new LinkedHashSet();
        Iterator expIt = this.expressions.iterator();
        while (expIt.hasNext()) {
            Expression exp = (Expression)expIt.next();
            Expression withVariants = insertVariants(exp,  tokenlists);
            if (withVariants!=null) result.add(withVariants);
            else ambiguous.add(exp);
        }
        return result;
    }
    
    Set getAmbiguousExpressions() {
        return this.ambiguous;
    }
    
    
    private Expression insertVariants(Expression expression, TokenList[] tokenlists) {
        //return true if variants have been inserted, false otherwise
        try {
            Map bindings = bindTokens2Nodes(expression,tokenlists);
            Map nodes2tokens = new IdentityHashMap();
            Util.reverseMapping(nodes2tokens, bindings);
            Map path2variants = findPathOfVariants(expression, nodes2tokens);
            return insertASTVariantNodes(expression, path2variants);
        }
        catch (BindingException bindexc) {
            log.fine("Ambiguous expression; detected mismatches: "+bindexc.getMismatches());
        }
        return null;
    }
    
    private Map bindTokens2Nodes(Expression expression, TokenList[] tokenlists) throws BindingException {
        Binder binder = new Binder(expression);
        Map binds = new IdentityHashMap();
        for(int i=0; i<tokenlists.length; i++) {
            binder.setBindings(tokenlists[i]);
            binds.putAll(binder.getTokenBindings());
        }
        return binds;
    }
    
    private Map findPathOfVariants(final Expression expression, final Map nodes2tokens) {
        final Map path2samples = new HashMap();
        new Walker(expression) {
            
            public boolean visit(ASTToken node) {
                if (node.maybeVariant()) {
                    List tokens = (List)nodes2tokens.get(node);
                    if (tokens!=null && tokens.size()>1) {
                        Set variants = new HashSet();
                        Iterator it = tokens.iterator();
                        while (it.hasNext()) {
                            Token token = (Token)it.next();
                            variants.add(token.getVariantValue());
                        }
                        if (variants.size()>1) { //hit variant
                            path2samples.put(getCurrentAbsolutePath(), variants);
                        }
                    }
                }
                return false;
            }
            
        }.visit(expression);
        return path2samples;
    }
    
    private Expression insertASTVariantNodes(Expression exp, Map path2samples) {
        Expression result = exp.clone(true); //modify a new copy of old expression
        ASTBuilder builder = new ASTBuilder();
        Iterator it = path2samples.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry)it.next();
            Node.Path path = (Node.Path)entry.getKey();
            Set samples = (Set)entry.getValue();
            Token token = (Token)exp.getNode(path);
            ASTVariant variant = builder.createVariant(token);
            variant.setSamples(samples);
            result.replaceNode(path, variant);
        }
        return result;
    }
}