/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * SearchSpace.java
 *
 * Created on 17 febbraio 2003, 18.06
 */

package roadrunner.engine.space;

import roadrunner.engine.ExpressionIterator;
import java.util.*;
import java.util.logging.*;

import roadrunner.bidi.Direction;
import roadrunner.ast.Expression;
import roadrunner.ast.ExpressionRegion;
import roadrunner.parser.Parser;
import roadrunner.parser.TokenList;
import roadrunner.parser.MismatchPoint;
import roadrunner.parser.AmbiguityMismatchFactory;
import roadrunner.engine.Backtracker;
import roadrunner.engine.space.operator.AbstractOperatorFactory;

/**
 *
 * @author  Valter Crescenzi
 */
public class SearchSpace extends Backtracker {
    
    static private Logger log = Logger.getLogger(SearchSpace.class.getName());
    
    /** Lo stato dell'allineamento si riduce ad una semplice exp. reg., ma servono anche mismatches ed operatori
     *  In questo spazio degli stati:      Operatore: stato -> { stato }
     */
    
/*  Lavorando  con le regioni piuttosto che le espressioni si hanno due vantaggi sostanziali:
 *   1) rende la classe piu' generale perche per passare da una exp ad una regione basta fare exp.asRegion();
 *   2) si evitano alcune materializzazioni che costano parecchio perche' implicano il cloning dei nodi delle exp
 */
        
    private Parser parser;
    
    private Direction dir;

    private ExpressionRegion region;
    private TokenList tokenlist;
    
    private AbstractOperatorFactory operatorFactory;
    private AmbiguityMismatchFactory mismatchFactory;
    
    private SpaceOptions options;
    
    private int numberOfStates;
    
    /** Creates a new instance of SearchSpace */
    public SearchSpace(Expression exp, TokenList tokenlist, int minDOMdepth) {
        this(exp.asRegion(),tokenlist,Direction.LEFT2RIGHT, minDOMdepth);
    }
    SearchSpace(Expression exp, TokenList tokenlist, Direction dir, int minDOMdepth) {
        this(exp.asRegion(),tokenlist,dir,minDOMdepth);
    }
    
    private SearchSpace(ExpressionRegion r, TokenList tl, Direction dir, int minDOMdepth) {
        super(new Expression(r));
        init(r,tl,dir,minDOMdepth);
        log.finest("Creating Search Space "+id+" "+dir+" to match with Tokenlist: "+tl);
    }
    
    private void init(ExpressionRegion r, TokenList tl, Direction dir, int minDOMdepth) {
        this.dir = dir;
        this.region = r;
        this.tokenlist = tl;
        this.options = new SpaceOptions(minDOMdepth);
        this.operatorFactory = AbstractOperatorFactory.getInstance(this);
        this.mismatchFactory = new AmbiguityMismatchFactory();
        this.numberOfStates = 0;
    }
    
    public SpaceOptions getOptions() {
        return this.options;
    }
    
    public AbstractOperatorFactory getOperatorFactory() {
        return this.operatorFactory;
    }
    
    TokenList getTokenlist() {
        return this.tokenlist;
    }
    
    protected ExpressionIterator successor(Expression current) {
        this.numberOfStates++;
        if (this.numberOfStates==this.options.getMaxNumberOfStates())
            return ExpressionIterator.expiredIterator();
        List mismatches = this.parser.getMismatches();
        if (options.getAmbiguity()!=0)
            addAmbiguityMismatches(mismatches);
        setSpaceOfMismatches(mismatches);
        State state = new State(mismatches, this);
        log.finest("Space "+id+" - Reached State "+state.getId());
        log.fine("Space "+id+" - Expression from State "+state.getId()+"\n"+current.dump("\t"));
        return state;
    }
    
    private void addAmbiguityMismatches(List mismatches) {
        int len = mismatches.size();
        for(int i=0; i<len; i++) {
            MismatchPoint m = (MismatchPoint)mismatches.get(i);
            List ambiguityMismatches = mismatchFactory.getAmbiguityMismatches(m);
            mismatches.addAll(ambiguityMismatches);
        }
    }
    
    private void setSpaceOfMismatches(List mismatches) {
        Iterator it = mismatches.iterator();
        while (it.hasNext()) {
            MismatchPoint m = (MismatchPoint)it.next();
            log.fine("Space "+id+" - mismatch "+m.getId()+":\t"+m);
            m.setSearchSpace(this);
        }
    }
    
    protected boolean goalTest(Expression exp) {
        this.parser = new Parser(exp,this.dir);
        return parser.parseAllExpression(this.tokenlist);
    }
    
    private int solutions = 0; //this fields and next two methods only for logging
    protected Expression gotFinal(Expression exp) {
        this.solutions++;
        log.fine("Space "+id+" - Found solution n. "+solutions);
        return exp;
    }
    
    protected void expired() {
        log.fine("Space "+id+" - No other solutions");
    }
    
    ExpressionRegion getMatchingRegion() {
        return this.parser.getMatchingRegion();
    }
    
    int getPosReachedOnTokenlist() {
        return this.parser.getPosReachedOnTokenlist();
    }
    
}
