/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
package roadrunner.engine.space.operator;

/**
 * Searcher.java
 *
 *
 * Created: Sun Nov 21 11:02:08 1999
 *
 * @author Valter Crescenzi
 * @version
 */

import java.util.*;

import roadrunner.bidi.*;
import roadrunner.ast.Node;
import roadrunner.ast.Matchable;
import roadrunner.ast.PrefixHunter;
import roadrunner.parser.Token;
import roadrunner.config.Config;
import roadrunner.config.Constants;

class Searcher {
    
    private BidirectionalList listView;
    private Region where;
    private Token wanted;
    private Direction dir;
    
    private List occurrencesIndices;
    private int numberOfOccurrences;
    
    
    Searcher(Region where, Token wanted, Direction dir) {
        this.listView = BidirectionalListFactory.newListView(dir, where.asList());
        this.where = where;
        this.wanted = wanted;
        this.dir = dir;
        this.occurrencesIndices =  new ArrayList(Math.min(getMaxOccurrencesAllowed(),5));
        findOccurrences();
    }
    

    private int getMaxOccurrencesAllowed() {
        int maxOcc = Config.getPrefs().getInt(Constants.MAX_NUM_OCCUR);;
        return (maxOcc>0 ? maxOcc : Integer.MAX_VALUE);
    }
    
    // Evitare di restituire regioni vuote per quanto detto di seguito:
    // skip the mismatch but leave it in the square
    // otherwise mismatches on aligning token could give empty squareTuple
    
    private void findOccurrences() {
        this.numberOfOccurrences = 0;
        int max = getMaxOccurrencesAllowed();
        
        ListIterator lit = this.listView.listIterator(dir);
        while (lit.hasNext() && this.numberOfOccurrences < max) {
            Object next = PrefixHunter.getFirstLeaf(dir, (Node)lit.next());
            
            if (!(next instanceof Matchable)) continue;
            
            Matchable cursor = (Matchable)next;
            
            if (cursor.depth()<this.wanted.depth()) break;  // too late: the "Matchable" wanted is deeper
            
            if (cursor.matches(this.wanted)) {
                this.occurrencesIndices.add(new Integer(lit.previousIndex()));
                this.numberOfOccurrences++;
            }
        }
    }
    
    
    int getNumberOfOccurrences() {
        return this.numberOfOccurrences;
    }
    
    Region getRegionToOccurrence(int noccurrence) {
        // we want the  occurrence to be in the square; see addPlus
        return dir.subRegionToNIndex(where, getOccurrenceIndex(noccurrence));
    }
    Region getRegionBeforeOccurrence(int noccurrence) {
        // we don't want the occurrence to be in the square; see addHook
        return dir.subRegionBeforeNIndex(where, getOccurrenceIndex(noccurrence));
    }
    Region getRegionAfterOccurrence(int noccurrence) {
        return dir.subRegionAfterNIndex(where, getOccurrenceIndex(noccurrence));
    }
    Region getRegionFromOccurrence(int noccurrence) {
        // used to avoid contiguous hooks othewise introduced by addHook_w
        return dir.subRegionFromNIndex(where, getOccurrenceIndex(noccurrence));
    }
    
    private int getOccurrenceIndex(int noccurrence) {
        return ((Integer)(this.occurrencesIndices.get(noccurrence))).intValue();
    }
    
    
} // Searcher
