/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * SquareOperatorFactory.java
 *
 * Created on 20 marzo 2003, 11.48
 * @author  Valter Crescenzi
 */

package roadrunner.engine.space.operator;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;

import roadrunner.bidi.Region;
import roadrunner.ast.Matchable;
import roadrunner.parser.MismatchPoint;
import roadrunner.parser.Token;
import roadrunner.engine.space.Operator;

abstract class SquareOperatorFactory extends OperatorFactory {
    
    private Searcher searcher;

    SquareOperatorFactory(MismatchPoint m) {
        super(m);
    }
    
    Searcher getSearcher() { return this.searcher; }
    
    public List createOperators() {
        if (getRegionToSearch().isEmpty()) return Collections.EMPTY_LIST;
        if (getTagWanted()==null) return Collections.EMPTY_LIST;
        
        this.searcher = new Searcher(getRegionToSearch(),getTagWanted(),m.getDir());
        int nOfOcc = searcher.getNumberOfOccurrences();
        List result = new ArrayList(nOfOcc);
        for(int i=0; i<nOfOcc; i++) {
            Region square = getSquare(i+1);
            if (!isWellFormed(square)) break;
            result.add(createOperator(i+1));
        }
        return result;
    }
    
    abstract Operator createOperator(int k);
    
    abstract Region getRegionToSearch();
    abstract Token getTagWanted();
    abstract Region getSquare(int k);
    
    private boolean isWellFormed(Region region) {
        if (region.size()==0) return true; // empty squares are well-formed
        
        Stack stack = new Stack(region.size()); // a small local stack
        Iterator it = region.asList().iterator();
        while (it.hasNext()) {
            
            Object next = it.next();
            if (! (next instanceof Matchable)) continue;
            
            Matchable c = (Matchable)next;
            if (c.code()==0) continue; // text doesn't matter, consider only tags
            
            if (c.code()<0) {
                if (stack.size()==0 || stack.peek()+c.code()!=0) return false;
                else stack.pop();       // close tag scope
            }
            else stack.push(c.code()); // save an open tag
        }
        return (stack.size()==0);
    }
    
    private class Stack {
        
        int stack[];
        int head;
        
        Stack(int size) {
            this.head = 0;
            this.stack = new int[size];
        }
        
        void push(int code) { this.stack[head++]=code;}
        int pop()           {
            int result  = stack[--head];
            stack[head] = 0;
            return result;
        }
        int peek()          { return stack[head-1];   }
        int size()          { return head;            }
    }
    
}