/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Lexer.java
 *
 * Created on
 * @author  Valter Crescenzi
 */
package roadrunner.parser;

import java.util.*;
import java.util.logging.*;

import java.io.Reader;
import java.io.IOException;

import roadrunner.util.Util;
import roadrunner.util.DOMLoader;
import roadrunner.config.Config;
import roadrunner.config.Preferences;
import roadrunner.config.Constants;
import roadrunner.parser.token.TokenFactory;
import roadrunner.labeller.Box;

import org.w3c.dom.Node;
import org.w3c.dom.Comment;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.xml.sax.SAXException;

public class Lexer implements Constants {
    
    static private Logger log = Logger.getLogger(Lexer.class.getName());
    
    private TokenFactory factory;
    private List tokens;
    private Reader in;
    private Preferences prefs;
    //labelling stuff
    private boolean labelling;
    private IdentityHashMap token2bb;
    private Box box;
    
    public Lexer(Reader in, Preferences prefs) {
        this.factory = TokenFactory.getInstance();
        this.in = in;
        this.prefs = prefs;
        this.tokens = null;
        this.labelling = Config.getPrefs().getBoolean(VISUAL_ENABLED);
        this.token2bb = new IdentityHashMap();
        this.box = null;
    }
    
    public List getTokens() throws SAXException, IOException  {
        this.tokens = new ArrayList();
        Document doc = DOMLoader.parseHTML(this.in);
        if (freetextEnabled()) {
            doc = new FreeTextHunter(doc, this.prefs).getCollapsedDocument();
        }
        getTokens(doc,0);
        return this.tokens;
    }
    
    public IdentityHashMap getTokens2BoundingBoxes() {
        return this.token2bb;
    }
    
    /* DOM Document -> Token List */
    private void getTokens(Node node,int depth) {
        int type = node.getNodeType();
        switch ( type ) {
            
            case Node.DOCUMENT_NODE:
                Document document = (Document)node;
                getTokens(document.getDocumentElement(),depth);
                break;
                
            case Node.ELEMENT_NODE:
                if (isSubtreeToSkip(node)) break;
                
                boolean toSkip = isTagToSkip(node);
                
                Token token = null;
                
                /* open tag */
                if (!toSkip) {
                    token = (Token)factory.createOpenTagToken(node,depth,this.prefs);
                    this.tokens.add(token);
                }
                /* visit subtrees */
                NodeList children = node.getChildNodes();
                if ( children != null ) {
                    int len = children.getLength();
                    for ( int i = 0; i < len; i++ ) {
                        getTokens(children.item(i), toSkip ? depth : depth+1);
                    }
                }
                /* close tag */
                if (!toSkip) {
                    this.tokens.add(factory.createCloseTagToken(token));
                }
                break;
                
            case Node.TEXT_NODE:
                // skip strings of all white spaces
                if (!Util.isAllWhiteSpaceChars(getText(node))) {
                    Token text = factory.createTextToken(node,depth);
                    this.tokens.add(text);
                    if (isVisualLabellingEnabled()) {
                        saveBoundingBoxOf(text);
                    }
                }
                break;
            case Node.COMMENT_NODE:
                String comment = node.getNodeValue();;
                
                if (Box.isBoundingBoxEncoding(comment) && isVisualLabellingEnabled()) {
                    this.box = Box.decode(comment); //Save parsed Box
                }
                break;
                /* skip anything else */
        }
    }
    
    private boolean isTagToSkip(Node element) {
        return this.prefs.getSet(IGNORE_TAGS).contains(getTagElement(element));
    }
    
    private boolean isSubtreeToSkip(Node element) {
        return this.prefs.getSet(IGNORE_TREES).contains(getTagElement(element));
    }
    
    private String getText(Node node) {
        return node.getNodeValue().trim();
    }
    
    private String getTagElement(Node node) {
        return ((Element)node).getNodeName().toLowerCase().trim();
    }
    
    private boolean freetextEnabled() {
        return this.prefs.getBoolean(FREE_TEXT);
    }
    
    private boolean isVisualLabellingEnabled() {
        return labelling;
    }
    
    private void saveBoundingBoxOf(Token token) {
        log.finest("Box of \""+token+ "\" is "+this.box);
        this.token2bb.put(token, this.box);
        this.box = null;
    }
    
    
}
